package de.communardo.confluence.plugins.displaysubspace;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import com.atlassian.confluence.spaces.Space;
import com.atlassian.confluence.spaces.SpaceType;

import de.communardo.confluence.plugins.createsubspace.CreateSubSpaceUtils;

/**
 * 
 * @author Bill Martin
 * 
 */
public class HierarchyBuilder {

    // To access a part of the hierarchy very quickly.
    private final HashMap<String, TreeNode> treeNodeCache;
    // The hierarchy.
    private final SpaceHierarchy spaceHierarchy;
    // The TreeNode Comparator
    private final TreeNodeComparator comparator;
    private final Map<String, Space> temporarySaved = new HashMap<String, Space>();
    private int hierarchyDepth;
    private Space parentSpace;
    private boolean showRoot;
    private final List<Space> spaces;
    private List<Space> permittedSpaces;

    /** Logger. */
    private static final Logger LOG = Logger.getLogger(HierarchyBuilder.class);

    public HierarchyBuilder(List<Space> spaces, String sort) {
        treeNodeCache = new HashMap<String, TreeNode>();
        comparator = new TreeNodeComparator(sort);
        spaceHierarchy = new SpaceHierarchy(comparator);
        this.setHierarchyDepth(-1);
        this.setParentSpace(null);
        this.spaces = spaces;

    }

    public void init(List permittedSpaces) {
        this.permittedSpaces = permittedSpaces;
        buildHierarchy(this.spaces);
    }

    /**
     * Build hierarchy.
     * 
     * @param spaces
     *            List with spaces.
     */
    private void buildHierarchy(List<Space> spaces) {
        Iterator<Space> iterator;
        int spacesSize;
        while (!spaces.isEmpty()) {
            spacesSize = spaces.size();
            iterator = spaces.iterator();
            while (iterator.hasNext()) {
                Space space = iterator.next();
                if (insert(space) || hasCyclicReference(space) || !parentIsGlobalSpace(space)) {
                    iterator.remove();
                }
            }
            // prevent infinite loop
            if (spaces.size() == spacesSize) {
                String spaceNames = "";
                for (Iterator<Space> iterator2 = spaces.iterator(); iterator2.hasNext();) {
                    spaceNames = spaceNames.concat(iterator2.next().getName());
                    if (iterator2.hasNext()) {
                        spaceNames = spaceNames.concat(", ");
                    }
                }
                LOG.debug("Unable to add space \"" + spaceNames + "\" in space hierarchy.");
                break;
            }
        }
    }

    /**
     * 
     * @return The sorted hierarchy.
     */
    public Collection<TreeNode> getHierarchy() {
        return spaceHierarchy.getHierarchy();
    }

    /**
     * Create a new TreeNode and store it in the cache and in the hierarchy.
     * 
     * @param space
     *            The Space
     * @return <code>true</code> if node was handled and can be removed from stack
     */
    private boolean insert(Space space) {

        Space parent = CreateSubSpaceUtils.getParentSpace(space);

        boolean permitted = this.permittedSpaces.contains(space);

        // temporary save not permittet spaces because of permitted childs
        if (!permitted) {
            temporarySaved.put(space.getKey(), space);
            return true;
        }
        if (parent != null && temporarySaved.get(parent.getKey()) != null) {
            reIntegrateTemps(parent);
        }

        boolean inserted = true;
        if (parent == null) {
            if (this.parentSpace == null
                    || (this.parentSpace != null && this.parentSpace == space && this.showRoot)) {
                inserted = insertSpace(space);
            }
        } else if (treeNodeCache.containsKey(parent.getKey())) {
            inserted = insertSpace(space, treeNodeCache.get(parent.getKey()));
        } else if (this.parentSpace != null
                && ((space.getId() == this.parentSpace.getId() && this.showRoot) || (parent.getId() == this.parentSpace
                        .getId() && !this.showRoot))) {
            inserted = insertSpace(space);
        } else {
            inserted = false;
        }
        return inserted;
    }

    private void reIntegrateTemps(Space parent) {
        Space parentOfParent = CreateSubSpaceUtils.getParentSpace(parent);
        if (parentOfParent != null && temporarySaved.get(parentOfParent.getKey()) != null) {
            reIntegrateTemps(parentOfParent);
        }
        if (parentOfParent == null) {
            insertSpace(temporarySaved.get(parent.getKey()));
        } else if (treeNodeCache.containsKey(parentOfParent.getKey())) {
            insertSpace(temporarySaved.get(parent.getKey()), treeNodeCache.get(parentOfParent
                    .getKey()));
        }
        temporarySaved.remove(parent.getKey());
    }

    /**
     * Decides whether given parent space is amongst children of given child space.
     * 
     * @param space
     * @return true if cyclic reference exists
     */
    private boolean hasCyclicReference(Space space) {
        boolean check = CreateSubSpaceUtils.hasCyclicReference(CreateSubSpaceUtils
                .getParentSpace(space), space);
        if (check) {
            LOG.warn("The space \"" + space.getName()
                    + "\" has a cycle reference and will be removed.");
        }
        return check;
    }

    /**
     * Decides whether the parent´s type of the given space is global.
     * 
     * @param space
     * @return true if space is global
     */
    private boolean parentIsGlobalSpace(Space space) {
        boolean check = SpaceType.isGlobal(CreateSubSpaceUtils.getParentSpace(space));
        if (!check) {
            LOG.debug("The space \"" + space.getName() + "\" references to a non global space.");
        }
        return check;
    }

    /**
     * 
     * @param space
     */
    private boolean insertSpace(Space space, TreeNode parent) {
        TreeNode node = new TreeNode(space, comparator, parent);
        int nodeLevel = node.getNodeLevel();
        if (this.hierarchyDepth > -1 && nodeLevel > this.hierarchyDepth) {
            return true;
        }
        treeNodeCache.put(space.getKey(), node);
        spaceHierarchy.insertNode(node, parent);
        return true;
    }

    /**
     * 
     * @param space
     */
    private boolean insertSpace(Space space) {
        return insertSpace(space, null);
    }

    public Space getParentSpace() {
        return parentSpace;
    }

    public int getHierarchyDepth() {
        return hierarchyDepth;
    }

    public void setHierarchyDepth(int hierarchyDepth) {
        this.hierarchyDepth = hierarchyDepth;
    }

    public void setParentSpace(Space parentSpace) {
        this.parentSpace = parentSpace;
    }

    public void setShowRoot(boolean withRoot) {
        this.showRoot = withRoot;
    }

    public boolean isShowRoot() {
        return showRoot;
    }

    public TreeNode getNodeFromKey(String currentSpace) {
        return treeNodeCache.get(currentSpace);
    }

}